__all__ = ["HasMapping"]


import asyncio
import pathlib
from typing import Dict, Any, Optional, List
from abc import ABC, abstractmethod

from yaqd_core import IsSensor, IsDaemon


class HasMapping(IsSensor, IsDaemon, ABC):
    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self.__mappings: Dict[str, Any] = (
            dict()
        )  # don't interact directly, use property
        self._mapping_id: int = 0
        self._channel_mappings: Dict[str, List[str]] = dict()
        self._mapping_units: Dict[str, str] = dict()
        self._mappings: Dict[str, object] = dict()

    @property
    def _mappings(self) -> Dict[str, Any]:
        return self.__mapped

    @_mappings.setter
    def _mappings(self, new):
        self.__mapped = new
        self._mapping_id += 1
        self.__mapped["mapping_id"] = self._mapping_id

    def get_channel_mappings(self) -> Dict[str, List[str]]:
        return self._channel_mappings

    def get_mapping_id(self) -> int:
        return self._mapping_id

    def get_mapping_units(self) -> Dict[str, str]:
        return self._mapping_units

    def get_mappings(self) -> Dict[str, object]:
        return self._mappings



__all__ = ["HasLimits"]


import pathlib
from typing import Dict, Any, Optional, List

from yaqd_core import HasPosition, IsDaemon


class HasLimits(HasPosition, IsDaemon):
    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._out_of_limits = config["out_of_limits"]

    def get_limits(self) -> List[float]:
        # wrapper for client
        return self.limits

    @property
    def limits(self) -> List[float]:
        # for internal use
        return self._joint_limit(self._state["hw_limits"], self._config["limits"])

    @classmethod
    def _joint_limit(cls, *limits: List[float]):
        mins, maxes = [*zip(*limits)]
        assert all([mini < maxi for mini, maxi in zip(mins, maxes)])
        out = [max(*mins), min(*maxes)]
        assert out[0] < out[1]
        return out

    def in_limits(self, position: float) -> bool:
        return self._in_limits(position)

    def _in_limits(self, position):
        # for internal use
        low, upp = self.limits
        return low <= position <= upp

    def set_position(self, position: float) -> None:
        if not self._in_limits(position):
            if self._out_of_limits == "closest":
                low, upp = self.limits
                if position > upp:
                    position = upp
                elif position < low:
                    position = low
            elif self._out_of_limits == "ignore":
                return
            else:
                raise ValueError(f"{position} not in ranges {self.limits}")
        super().set_position(position)



__all__ = ["Sensor"]


from yaqd_core import HasMeasureTrigger, IsSensor, IsDaemon


class Sensor(HasMeasureTrigger, IsSensor, IsDaemon):
    """Legacy class left for downstream packages that expect it. Do not use in new designs."""

    pass



from dataclasses import dataclass, asdict
from enum import Enum
import json
from io import BytesIO
from typing import Dict, Optional

import fastavro  # type: ignore

from .protocol import hash_protocol

handshake_request_schema = fastavro.parse_schema(
    {
        "type": "record",
        "name": "HandshakeRequest",
        "namespace": "org.apache.avro.ipc",
        "fields": [
            {
                "name": "clientHash",
                "type": {"type": "fixed", "name": "MD5", "size": 16},
            },
            {"name": "clientProtocol", "type": ["null", "string"]},
            {"name": "serverHash", "type": "MD5"},
            {"name": "meta", "type": ["null", {"type": "map", "values": "bytes"}]},
        ],
    }
)


@dataclass
class HandshakeRequest:
    """avro RPC handshake request."""

    clientHash: bytes
    clientProtocol: Optional[str]
    serverHash: bytes
    meta: Optional[Dict[str, bytes]] = None

    def __bytes__(self):
        fo = BytesIO()
        fastavro.schemaless_writer(fo, handshake_request_schema, asdict(self))
        fo.seek(0)
        return fo.read()

    @classmethod
    def from_bytes(cls, data):
        fo = BytesIO()
        fo.write(data)
        fo.seek(0)
        return cls(**fastavro.schemaless_reader(fo, handshake_request_schema))


class HandshakeMatch(str, Enum):
    BOTH = "BOTH"
    CLIENT = "CLIENT"
    NONE = "NONE"


handshake_response_schema = fastavro.parse_schema(
    {
        "type": "record",
        "name": "HandshakeResponse",
        "namespace": "org.apache.avro.ipc",
        "fields": [
            {
                "name": "match",
                "type": {
                    "type": "enum",
                    "name": "HandshakeMatch",
                    "symbols": ["BOTH", "CLIENT", "NONE"],
                },
            },
            {"name": "serverProtocol", "type": ["null", "string"]},
            {
                "name": "serverHash",
                "type": ["null", {"type": "fixed", "name": "MD5", "size": 16}],
            },
            {"name": "meta", "type": ["null", {"type": "map", "values": "bytes"}]},
        ],
    }
)


@dataclass
class HandshakeResponse:
    match: HandshakeMatch
    serverProtocol: Optional[str]
    serverHash: Optional[bytes]
    meta: Optional[Dict[str, bytes]] = None

    def __dict__(self):
        asdict(self)

    def __bytes__(self):
        fo = BytesIO()
        fastavro.schemaless_writer(fo, handshake_response_schema, asdict(self))
        fo.seek(0)
        return fo.read()

    @classmethod
    def from_bytes(cls, data):
        fo = BytesIO()
        fo.write(data)
        fo.seek(0)
        return cls(**fastavro.schemaless_reader(fo, handshake_request_schema))


def handle_handshake(request, protocol):
    request = HandshakeRequest(**request)
    # TODO actual caching

    match = HandshakeMatch.NONE
    server_protocol = None
    server_hash = None

    if request.clientProtocol is None:
        match = HandshakeMatch.NONE
    elif request.serverHash == hash_protocol(protocol):
        match = HandshakeMatch.BOTH

    if match != HandshakeMatch.BOTH:
        server_protocol = json.dumps(protocol)
        server_hash = hash_protocol(protocol)
    return HandshakeResponse(
        match=match, serverProtocol=server_protocol, serverHash=server_hash
    )



__all__ = ["ASerial", "get_aserial"]

import asyncio
from typing import Any, Dict

import serial  # type: ignore


class ASerial(serial.Serial):
    def __init__(self, port=None, baudrate=9600, eol=b"\n", *args, **kwargs):
        super().__init__(port, baudrate, timeout=0, *args, **kwargs)
        self._readlock = asyncio.Lock()
        self.eol = eol

    async def aread(self, size=1):
        async with self._readlock:
            return await self._aread(size)

    async def _aread(self, size=1):
        buf = b""
        while len(buf) < size:
            buf += self.read(size - len(buf))
            await asyncio.sleep(0.01)
        return buf

    async def areadline(self, size=-1):
        async with self._readlock:
            return await self._areadline(size)

    async def _areadline(self, size=-1):
        buf = b""
        while not buf.endswith(self.eol) and (size < 0 or len(buf) < size):
            buf += self.readline(size - len(buf) if size >= 0 else -1)
            await asyncio.sleep(0.01)
        return buf

    async def areadlines(self, hint=-1):
        async with self._readlock:
            size = 0
            while hint < 0 or size < hint:
                s = await self._areadline()
                size += len(s)
                yield s

    async def awrite_then_read(self, data, size=1):
        async with self._readlock:
            self.write(data)
            return await self._aread(size)

    async def awrite_then_readline(self, data, size=-1):
        async with self._readlock:
            self.write(data)
            return await self._areadline(size)


_serial_objects: Dict[str, ASerial] = {}


def get_aserial(
    port: str,
    baudrate: int = 9600,
    eol: bytes = b"\n",
    **kwargs: Any,
) -> ASerial:
    """Create a new ASerial object or return already existed one.

    Parameters
    ----------
    port: str
        Serial port identificator, e.g. 'COM0' or '/dev/ttyUSB0'
        If an matching open port exists, all other arguments are ignored.
    baudrate: int, optional
        Baud rate of the port. Defaults to 9600.
    eol: bytes, optional
        End of line terminator. Defaults to new-line.

    Other Parameters
    ----------------
    **kwargs: `serial.Serial` arguments, optional


    Returns
    -------
    ASerial
        Corresponding ASerial object
    """
    if port not in _serial_objects:
        _serial_objects[port] = ASerial(port=port, baudrate=baudrate, eol=eol, **kwargs)
    return _serial_objects[port]



"""Special class for saving state. A dictonary with attribute 'updated'."""


class State(dict):
    def __init__(self, *args, **kwargs):
        """A dictionary with special attribute 'updated'.
        Internal behavior will never set updated to False.
        """
        super().__init__(*args, **kwargs)
        self.updated = True

    def __setitem__(self, item, value):
        super().__setitem__(item, value)
        self.updated = True

    def update(self, iterable):
        super().update(iterable)
        self.updated = True



import sys
import subprocess
import time


__all__ = ["run_daemon_entry_point", "run_daemon_from_file"]


def run_daemon_entry_point(kind, config):
    def decorator(function):
        def wrapper():
            with subprocess.Popen([f"yaqd-{kind}", "--config", config]) as proc:
                tries = 100
                while True:
                    # Process exited with nonzero exit status
                    if proc.poll():
                        raise subprocess.SubprocessError(proc.stderr)
                    if tries <= 0:
                        function()
                    try:
                        function()
                    except ConnectionError:
                        time.sleep(0.1)
                    except:
                        proc.terminate()
                        raise
                    else:
                        break
                    tries -= 1
                proc.terminate()

        return wrapper

    return decorator


def run_daemon_from_file(pyfile, config):
    def decorator(function):
        def wrapper():
            with subprocess.Popen([sys.executable, pyfile, "--config", config]) as proc:
                tries = 100
                while True:
                    # Process exited with nonzero exit status
                    if proc.poll():
                        raise subprocess.SubprocessError(proc.stderr)
                    if tries <= 0:
                        function()
                    try:
                        function()
                    except ConnectionError:
                        time.sleep(0.1)
                    except:
                        proc.terminate()
                        raise
                    else:
                        break
                    tries -= 1
                proc.terminate()

        return wrapper

    return decorator



__all__ = ["IsDiscrete"]


import pathlib
from typing import Dict, Any, Optional, List


import yaqd_core


class IsDiscrete(yaqd_core.HasPosition, yaqd_core.IsDaemon):
    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        self._position_identifiers: Dict[str, float] = config.get("identifiers", {})
        super().__init__(name, config, config_filepath)

    def get_position_identifiers(self) -> Dict[str, float]:
        return self._position_identifiers

    def get_position_identifier_options(self) -> List[str]:
        return list(self._position_identifiers.keys())

    def set_identifier(self, identifier: str) -> float:
        p = self._position_identifiers[identifier]
        self.set_position(p)
        return p

    def get_identifier(self) -> str:
        return self._state["position_identifier"]



from .unpacker import *
from . import avro_numpy

import fastavro  # type: ignore
from io import BytesIO


def fill_avro_default(schema, partial, named_types=None):
    try:
        schema = fastavro.parse_schema(
            schema["type"], expand=True, named_schemas=named_types
        )
    except fastavro.schema.UnknownType:
        # e.g. array with flattend items types
        schema = fastavro.parse_schema(schema, expand=True, named_schemas=named_types)

    # Sometimes fastavro.parse_schema doesn't add these like I think it should
    # One way that works with some schemas is to run it through fastavro twice.
    # However, that does not work if a map or array is the top level object.
    # Later on, when consuming the schema (read or write) it still looks for these
    # special internal behavior keys, so this just directly adds them indiscriminantly.
    # In doing so, fastavro can use the named_schemas as a lookup table whenever
    # it needs one of the named types.
    # It is still necessary to do the first pass for reasons that are not fully
    # clear, but in part because it does expand the string type name into the dict
    # in which these are placed.
    if isinstance(schema, dict):
        schema["__fastavro_parsed"] = True
        schema["__named_schemas"] = named_types

    s = BytesIO()
    fastavro.schemaless_writer(s, schema, partial)
    s.seek(0)
    return fastavro.schemaless_reader(s, schema)



__all__ = ["ContinuousHardware"]


from yaqd_core import HasLimits, HasPosition, IsDaemon


class ContinuousHardware(HasLimits, HasPosition, IsDaemon):
    """Legacy class left for downstream packages that expect it. Do not use in new designs."""

    pass



__all__ = ["HasDependents"]


from abc import abstractmethod
from typing import Dict

import yaqd_core


class HasDependents(yaqd_core.IsDaemon):
    @abstractmethod
    def get_dependent_hardware(self) -> Dict[str, str]:
        raise NotImplementedError



__all__ = ["IsHomeable"]


from abc import abstractmethod

import yaqd_core


class IsHomeable(yaqd_core.HasPosition, yaqd_core.IsDaemon):
    @abstractmethod
    def home(self):
        raise NotImplementedError



import yaqd_core


import copy


def assert_mro(cls, avpr):
    traits = list(copy.copy(avpr["traits"]))  # no particular order
    mro = list(cls.__mro__)
    while mro:
        cls = mro.pop(0)
        if cls == yaqd_core.HasTransformedPosition:
            assert traits.pop(traits.index("has-transformed-position"))
            assert yaqd_core.HasLimits in mro
            assert yaqd_core.HasPosition in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasLimits:
            assert traits.pop(traits.index("has-limits"))
            assert yaqd_core.HasPosition in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasDependents:
            assert traits.pop(traits.index("has-dependents"))
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasMapping:
            assert traits.pop(traits.index("has-mapping"))
            assert yaqd_core.IsSensor in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasMeasureTrigger:
            assert traits.pop(traits.index("has-measure-trigger"))
            assert yaqd_core.IsSensor in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasPosition:
            assert traits.pop(traits.index("has-position"))
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.HasTurret:
            assert traits.pop(traits.index("has-turret"))
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.IsDaemon:
            assert traits.pop(traits.index("is-daemon"))
        elif cls == yaqd_core.IsDiscrete:
            assert traits.pop(traits.index("is-discrete"))
            assert yaqd_core.HasPosition in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.IsHomeable:
            assert traits.pop(traits.index("is-homeable"))
            assert yaqd_core.HasPosition in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.IsSensor:
            assert traits.pop(traits.index("is-sensor"))
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.UsesSerial:
            assert traits.pop(traits.index("uses-serial"))
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.UsesI2C:
            assert traits.pop(traits.index("uses-i2c"))
            assert yaqd_core.UsesSerial in mro
            assert yaqd_core.IsDaemon in mro
        elif cls == yaqd_core.UsesUart:
            assert traits.pop(traits.index("uses-uart"))
            assert yaqd_core.UsesSerial in mro
            assert yaqd_core.IsDaemon in mro
        else:
            pass
    if traits:  # should be empty, will enter this case if not
        raise Exception(f"I didn't find a class matching trait(s): {traits}")



__all__ = ["HasTurret"]


from abc import abstractmethod

import yaqd_core


class HasTurret(yaqd_core.IsDaemon):
    @abstractmethod
    def set_turret(self, identifier: str):
        self._busy = True

    @abstractmethod
    def get_turret_options(self): ...

    def get_turret(self):
        return self._state["turret"]



from __future__ import annotations

__all__ = ["HasMeasureTrigger"]


import asyncio
import pathlib
from typing import Any
from abc import ABC, abstractmethod

from yaqd_core import IsSensor, IsDaemon


class HasMeasureTrigger(IsSensor, IsDaemon, ABC):
    def __init__(
        self, name: str, config: dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._looping = False
        if self._config["loop_at_startup"]:
            self.measure(loop=True)

    def get_measured(self) -> dict:
        return super().get_measured()

    @abstractmethod
    async def _measure(self) -> dict:
        """Do measurement, filling _measured dictionary.

        Returns dictionary with keys channel names, values numbers or arrays.
        """
        raise NotImplementedError

    def measure(self, loop: bool = False) -> int:
        """Start a measurement, optionally looping.

        Sensor will remain busy until measurement completes.

        Parameters
        ----------
        loop: bool, optional
            Toggle looping behavior. Default False.

        See Also
        --------
        stop_looping
        """
        self._looping = loop
        if not self._busy:
            self._busy = True
            self._tasks.append(self._loop.create_task(self._runner()))
        return self._measurement_id + 1

    async def _runner(self) -> None:
        """Handle execution of _measure, including looping and setting of _measurement_id."""
        while True:
            self._measured = await self._measure()
            assert set(self._measured.keys()) == set(self._channel_names)
            self._measurement_id += 1
            self._measured["measurement_id"] = self._measurement_id
            if "has-mapping" in self._traits:
                self._measured["mapping_id"] = self._mapping_id  # type: ignore
            if not self._looping:
                self._busy = False
                break
            await asyncio.sleep(0)
        current_task = asyncio.current_task()
        if current_task:
            self._tasks.remove(current_task)

    def stop_looping(self) -> None:
        """Stop looping."""
        self._looping = False



__all__ = ["HasTransformedPosition"]


import pathlib
from typing import Dict, Any, Optional, List
from yaqd_core import HasLimits, HasPosition, IsDaemon


class HasTransformedPosition(HasLimits, HasPosition, IsDaemon):
    def __init__(
        self, name: str, config: Dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._native_units = None

    # --- conversion of coordinates ---------------------------------------------------------------

    def to_native(self, transformed_position):
        relative = self._transformed_to_relative(transformed_position)
        return relative + self._state["native_reference_position"]

    def to_transformed(self, native_position):
        relative = native_position - self._state["native_reference_position"]
        return self._relative_to_transformed(relative)

    def _relative_to_transformed(self, relative_position):
        """convert a relative coordinate to a transformed coordinate.
        Relative coordinates differ from natural coordinates in that the null position has been subtracted.
        (i.e. in relative coordinates reference position is zero).

        This placeholder function has trivial behavior, f(x) = x.
        Daemons that need more than a reference_position should override this method.
        Note that override should still obey inversion:
        ```
        _relative_to_transformed(_transformed_to_relative(x)) == x
        ```.
        """
        return relative_position

    def _transformed_to_relative(self, transformed_position):
        """convert a transformed coordinate to a relative coordinate
        Relative coordinates differ from natural coordinates in that the null position has been subtracted.
        (i.e. in relative coordinates reference position is zero).

        This placeholder function has trivial behavior, f(x) = x.
        Daemons that need more than a reference_position should override this method.
        Note that override should still obey inversion:
        ```
        _relative_to_transformed(_transformed_to_relative(x)) == x
        ```.
        """
        return transformed_position

    # --- wrap parent messages that deal with position --------------------------------------------
    # --- clients use transformed position, but parent messages except/return native position.

    def set_position(self, position: float) -> None:
        super().set_position(self.to_native(position))

    def set_relative(self, distance: float) -> float:
        new = self.to_transformed(self._state["destination"]) + distance
        self.set_position(new)
        return new

    def get_position(self) -> float:
        position = super().get_position()
        return self.to_transformed(position)

    def get_destination(self) -> float:
        return self.to_transformed(super().get_destination())

    def in_limits(self, position: float) -> bool:
        return super().in_limits(self.to_native(position))

    def get_limits(self) -> List[float]:
        return sorted(map(self.to_transformed, self.limits))

    # --- setting or returning native coordinates -------------------------------------------------
    # --- new messages introduce by this trait

    @property
    def limits(self) -> List[float]:
        return self._joint_limit(
            self._state["hw_limits"],
            sorted(map(self.to_native, self._config["limits"])),
            self._config["native_limits"],
        )

    def get_native_reference(self) -> float:
        return self._state["native_reference_position"]

    def set_native_reference(self, native_position):
        self._state["native_reference_position"] = native_position

    def set_native_position(self, native_position):
        super().set_position(native_position)

    def get_native_position(self) -> float:
        return self._state["position"]

    def get_native_destination(self) -> float:
        return self._state["destination"]

    def get_native_limits(self) -> List[float]:
        return self.limits

    def get_native_units(self) -> Optional[str]:
        return self._native_units



__all__ = ["UsesSerial"]


from abc import abstractmethod

import yaqd_core


class UsesSerial(yaqd_core.IsDaemon):
    @abstractmethod
    def direct_serial_write(self, message: bytes):
        raise NotImplementedError



"""Core python package for implementing yaq deamons, and associated utilities."""

from . import logging
from .__version__ import __version__

from ._is_daemon import *
from ._has_position import *
from ._is_sensor import *
from ._has_dependents import *
from ._has_measure_trigger import *
from ._uses_serial import *
from ._uses_i2c import *
from ._uses_uart import *
from ._has_turret import *
from ._is_homeable import *
from ._is_discrete import *
from ._has_limits import *
from ._has_transformed_position import *
from ._has_mapping import *
from ._legacy import *

# need to define __all__ for pdoc, avoid having to manually add to it
__all__ = list(filter(lambda x: not x.startswith("_"), dir()))



"""Define version."""

import pathlib
import subprocess


here = pathlib.Path(__file__).resolve().parent

__all__ = ["__version__", "__branch__", "__avro_version__"]

__avro_version__ = "1.9.2"

__version__ = "2025.12.0"

try:
    __branch__ = (
        subprocess.run(
            ["git", "branch", "--show-current"], capture_output=True, cwd=here
        )
        .stdout.strip()
        .decode()
    )
except:
    __branch__ = ""

if __branch__:
    __version__ += "+" + __branch__



__all__ = ["Unpacker"]

import asyncio
import io
import struct

import fastavro  # type: ignore

from .handshake import handle_handshake, handshake_request_schema


class Unpacker:
    def __init__(self, protocol, file_like=None):
        self.protocol = protocol
        if file_like is None:
            self._file = io.BytesIO()
        else:
            self._file = file_like
        self.buf = io.BytesIO()
        self.handshake_complete = False
        self.handshake_response = None
        self.meta = None
        self.message_name = None
        self.parameters = None
        self.remaining = 0
        self.named_types = {t["name"]: t for t in self.protocol.get("types", [])}
        self.new_data = asyncio.Event()

    def __aiter__(self):
        return self

    async def __anext__(self):
        while True:
            try:
                self.new_data.clear()
                if not self.handshake_complete and self.handshake_response is None:
                    handshake_request = await self._read_object(
                        handshake_request_schema
                    )
                    self.handshake_response = handle_handshake(
                        handshake_request, self.protocol
                    )
                    if self.handshake_response.match == "BOTH":
                        self.handshake_complete = True
                if self.meta is None:
                    self.meta = await self._read_object(
                        {"type": "map", "values": "bytes"}
                    )
                if self.message_name is None:
                    self.message_name = await self._read_object("string")
                if self.message_name != "" and self.protocol["messages"][
                    self.message_name
                ].get("request", []):
                    await self._read_parameters(self.message_name)

                ret = (
                    self.handshake_response,
                    self.meta,
                    self.message_name,
                    self.parameters,
                )
                self.handshake_response = None
                self.meta = None
                self.message_name = None
                self.parameters = None
                return ret
            except (ValueError, struct.error):
                await self.new_data.wait()

    def feed(self, data: bytes):
        # Must support random access, if it does not, must be fed externally (e.g. TCP)
        pos = self._file.tell()
        self._file.seek(0, 2)
        if pos == self._file.tell():
            # read reached EOF and we are safe to clear
            self._file.seek(0)
            self._file.truncate(0)
            pos = 0
        self._file.write(data)
        self._file.seek(pos)
        self.new_data.set()

    async def _read_object(self, schema):
        schema = fastavro.parse_schema(
            schema, expand=True, named_schemas=self.named_types
        )
        try:
            # Needed twice for nested types... Should likely be fixed upstream
            schema = fastavro.parse_schema(
                schema, expand=True, named_schemas=self.named_types
            )
        except fastavro.schema.SchemaParseException:
            pass  # Must not have needed the second pass...
        while True:
            try:
                self.buf.seek(0)
                obj = fastavro.schemaless_reader(self.buf, schema)
                self.buf = io.BytesIO()
                return obj
            except Exception:
                self.buf.seek(0)
            if not self.remaining:
                self.remaining = struct.unpack_from(">L", self._file.read(4))[0]

            self.buf.seek(0, 2)
            num_read = self.buf.write(self._file.read(self.remaining))
            self.remaining -= num_read
            await asyncio.sleep(0)

    async def _read_parameters(self, name):
        if self.parameters is None:
            self.parameters = []
        for param_schema in self.protocol["messages"][name]["request"][
            len(self.parameters) :
        ]:
            self.parameters.append(await self._read_object(param_schema["type"]))



from __future__ import annotations

__all__ = ["IsSensor"]


import asyncio
import pathlib
from typing import Any

import yaqd_core


class IsSensor(yaqd_core.IsDaemon):
    def __init__(
        self, name: str, config: dict[str, Any], config_filepath: pathlib.Path
    ):
        super().__init__(name, config, config_filepath)
        self._measured: dict = dict()  # values must be numbers or arrays
        self._channel_names: list[str] = []
        self._channel_units: dict[str, str] = dict()
        self._channel_shapes: dict[str, tuple[int, ...]] = dict()
        self._measurement_id = 0
        self._measured["measurement_id"] = self._measurement_id

    def get_channel_names(self):
        """Get current channel names."""
        return self._channel_names

    def get_channel_shapes(self):
        """Get channel shapes."""
        # as default behavior, assume all channels are scalars
        if self._channel_shapes:
            return self._channel_shapes
        return {k: () for k in self._channel_names}

    def get_channel_units(self):
        """Get channel units."""
        return self._channel_units

    def get_measured(self) -> dict:
        assert "measurement_id" in self._measured
        return self._measured

    def get_measurement_id(self) -> int:
        return int(self._measured["measurement_id"])



import asyncio
import io
import struct
import traceback

import fastavro  # type: ignore
from . import avrorpc


class Protocol(asyncio.Protocol):
    def __init__(self, daemon, *args, **kwargs):
        self._daemon = daemon
        self.logger = daemon.logger
        self._avro_protocol = daemon._avro_protocol
        self._named_types = {t["name"]: t for t in self._avro_protocol.get("types", [])}
        asyncio.Protocol.__init__(self, *args, **kwargs)

    def connection_lost(self, exc):
        peername = self.transport.get_extra_info("peername")
        self.logger.info(f"Connection lost from {peername} to {self._daemon.name}")
        self.task.cancel()
        self._daemon._connection_lost(peername)

    def connection_made(self, transport):
        """Process an incomming connection."""
        peername = transport.get_extra_info("peername")
        self.logger.info(f"Connection made from {peername} to {self._daemon.name}")
        self.transport = transport
        self.unpacker = avrorpc.Unpacker(self._avro_protocol)
        self._daemon._connection_made(peername)
        self.task = asyncio.get_running_loop().create_task(self.process_requests())

    def data_received(self, data):
        """Process an incomming request."""
        self.logger.debug(f"Data received: {repr(data)}")
        if not self._daemon._server.is_serving():
            self.transport.close()
        self.unpacker.feed(data)

    async def process_requests(self):
        async for hs, meta, name, params in self.unpacker:
            if hs is not None:
                out = bytes(hs)
                out = struct.pack(">L", len(out)) + out
                self.transport.write(out)
                if hs.match == "NONE":
                    name = ""

            out_meta = io.BytesIO()
            fastavro.schemaless_writer(
                out_meta, {"type": "map", "values": "bytes"}, meta
            )
            length = out_meta.tell()
            self.transport.write(struct.pack(">L", length) + out_meta.getvalue())
            self.logger.debug(f"Wrote meta, {meta}, {out_meta.getvalue()}")
            try:
                response_out = io.BytesIO()
                response = None
                response_schema = "null"
                if name:
                    fun = getattr(self._daemon, name)
                    if params is None:
                        params = []
                    response = fun(*params)
                    response_schema = fastavro.parse_schema(
                        self._avro_protocol["messages"][name].get("response", "null"),
                        expand=True,
                        named_schemas=self._named_types,
                    )
                    # Needed twice for nested types... Probably can be fixed upstream
                    response_schema = fastavro.parse_schema(
                        response_schema,
                        expand=True,
                        named_schemas=self._named_types,
                    )
                fastavro.schemaless_writer(response_out, response_schema, response)
            except Exception as e:
                self.logger.error(f"Caught exception: {type(e)} in message {name}")
                self.logger.debug(traceback.format_exc())
                self.transport.write(struct.pack(">L", 1) + b"\1")
                error_out = io.BytesIO()
                fastavro.schemaless_writer(error_out, ["string"], repr(e))
                length = error_out.tell()
                self.transport.write(struct.pack(">L", length) + error_out.getvalue())
            else:
                self.transport.write(struct.pack(">L", 1) + b"\0")
                self.logger.debug(f"Wrote non-error flag")
                length = response_out.tell()
                self.transport.write(
                    struct.pack(">L", length) + response_out.getvalue()
                )
                self.logger.debug(
                    f"Wrote response {response}, {response_out.getvalue()}"
                )
            self.transport.write(struct.pack(">L", 0))
            if name == "shutdown":
                self.logger.debug("Closing transport")
                self.transport.close()
